--[[****************************************************************
	LootCount DropCount v0.20

	Author: Evil Duck
	****************************************************************

	Plug-in for the World of Warcraft add-on LootCount.

	****************************************************************]]

-- 0.20 Added support for sharing data in guild, some bugfixes
-- 0.10 First version


LOOTCOUNT_DROPCOUNT_VERSIONTEXT = "LootCount DropCount v0.20";
LOOTCOUNT_DROPCOUNT = "LootCount DropCount";
SLASH_DROPCOUNT1 = "/dropcount";
SLASH_DROPCOUNT2 = "/lcdc";
local LootCount_DropCount_Debug=nil;
local LootCount_DropCount_Registered=nil;
local LootCount_DropCount_Fighting=nil;
local LootCount_DropCount_Target=nil;
local LootCount_DropCount_GUID="0x0";
local LootCount_DropCount_SkinTarget=nil;
local LootCount_DropCount_SkinTargetKill=nil;
local LootCount_DropCount_LootedMobs={};
local LootCount_DropCount_SkinnedMobs={};
local LootCount_DropCount_ReceivedSkins={
	A="none",
	B="none",
	C="none",
};
local LCDC_LOOTEDAGE=1200;
-- Saved per character
LootCount_DropCount_Character={
	Skinning=nil,
	ShowZone=true,
};
-- Global save
LootCount_DropCount_DB={
	Item={},
	Count={},
};
LootCount_DropCount_NoQuest={
	[4582]=true,		-- Soft Bushy Tails
	[5113]=true,		-- Mark of the Syndicate
	[8483]=true,		-- Wastewander Pouch
};

local LCDC_LISTLENGTH=25;
local LCDC_KILLS_HIDE=10;
local LCDC_KILLS_SAFE=50;

local LootCount_DropCount_LootList={};
local LootCount_DropCount_TimedQueue = {};
local LootCount_DropCount_QueueSize=0;
local LCDC_QUEUETIME=900;
local LCT_QUEST=nil;
local LootCount_DropCount_EnumerateTimer=6;
local LootCount_DropCount_CacheRetries=0;
local LCSV_MAXRETRIES=10;
local LCR_QUEST=-1;
local LootCount_DropCount_This=nil;
local LCDC_PREFIX="LcDc";
local LCDC_SEPARATOR=",";
local LCDC_MOBKILL="MOB_KILL";
local LCDC_MOBLOOT="MOB_LOOT";
local LootCount_DropCount_UnknownItems=nil;
local LootCount_DropCount_RunningItem=nil;
local LootCount_DropCount_LastID=0;
local LootCount_DropCount_LastLootID=nil;
local LootCount_DropCount_Incoming={};
local LootCount_DropCount_SkinningList={
	[2934]=true,		-- Ruined Leather Scraps
	[2318]=true,		-- Light Leather
	[2319]=true,		-- Medium Leather
	[4234]=true,		-- Heavy Leather
	[4304]=true,		-- Thick Leather
	[8170]=true,		-- Rugged Leather
	[25649]=true,		-- Knothide Leather Scraps
	[21887]=true,		-- Knothide Leather
	[5082]=true,		-- Thin Kodo Leather
	[15423]=true,		-- Chimera Leather
	[15417]=true,		-- Devilsaur Leather
	[15422]=true,		-- Frostsaber Leather
	[15419]=true,		-- Warbear Leather
	[25703]=true,		-- Zhevra Leather
	[19767]=true,		-- Primal Bat Leather
	[19768]=true,		-- Primal Tiger Leather
	[17012]=true,		-- Core Leather
	[25699]=true,		-- Crystal Infused Leather
	[25708]=true,		-- Thick Clefthoof Leather
	[25700]=true,		-- Fel Scales
	[29539]=true,		-- Cobra Scales
	[29548]=true,		-- Nether Dragonscales
	[783]=true,			-- Light Hide
	[4232]=true,		-- Medium Hide
	[4235]=true,		-- Heavy Hide
	[8169]=true,		-- Thick Hide
	[8171]=true,		-- Rugged Hide
	[25707]=true,		-- Fel Hide
	[7428]=true,		-- Shadowcat Hide
	[8368]=true,		-- Thick Wolfhide
	[12731]=true,		-- Pristine Hide of the Beast
	[20500]=true,		-- Light Silithid Carapace
	[20501]=true,		-- Heavy Silithid Carapace
	[8167]=true,		-- Turtle Scale
	[8154]=true,		-- Scorpid Scale
	[15408]=true,		-- Heavy Scorpid Scale
	[20498]=true,		-- Silithid Chitin
	[29547]=true,		-- Wind Scales
	[7072]=true,		-- Naga Scale
	[7286]=true,		-- Black Whelp Scale
	[7287]=true,		-- Red Whelp Scale
	[7392]=true,		-- Green Whelp Scale
	[12607]=true,		-- Brilliant Chromatic Scale
	[8165]=true,		-- Worn Dragonscale
	[15412]=true,		-- Green Dragonscale
	[15415]=true,		-- Blue Dragonscale
	[15416]=true,		-- Black Dragonscale
	[15414]=true,		-- Red Dragonscale
	[15410]=true,		-- Scale of Onyxia
	[20381]=true,		-- Dreamscale
	[6470]=true,		-- Deviate Scale (drop and skinning)
	[6471]=true,		-- Perfect Deviate Scale (drop and skinning)
	[33567]=true,		-- Borean Leather Scraps
	[33568]=true,		-- Borean Leather

--	[12607]=true,		-- Brilliant Chromatic Scale 
};



-- Set up for handling
function LootCount_DropCount_OnLoad()
	this:RegisterEvent("VARIABLES_LOADED");
	this:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED");
	this:RegisterEvent("PLAYER_FOCUS_CHANGED");
	this:RegisterEvent("PLAYER_TARGET_CHANGED");
	this:RegisterEvent("CHAT_MSG_ADDON");
	this:RegisterEvent("CHAT_MSG_CHANNEL");
	this:RegisterEvent("LOOT_OPENED");
	this:RegisterEvent("LOOT_SLOT_CLEARED");

	SlashCmdList["DROPCOUNT"]=function(msg) LootCount_DropCount_Slasher(msg) end;
end


-- There's slashing to be done
function LootCount_DropCount_Slasher(msg)
-- Validate input
	if (not msg) then msg=""; end
	if (strlen(msg)>0) then msg=strlower(msg); end

	if (msg=="guild") then
		LootCount_DropCount_DB.GUILD=true;
		LootCount_DropCount_DB.RAID=nil;
		LootCount_Chat("DropCount: Sharing data with guild");
		return;
	elseif (msg=="raid" or msg=="group") then
		LootCount_DropCount_DB.GUILD=nil;
		LootCount_DropCount_DB.RAID=true;
		LootCount_Chat("DropCount: Sharing data with party/raid");
		return;
	elseif (msg=="noshare") then
		LootCount_DropCount_DB.GUILD=nil;
		LootCount_DropCount_DB.RAID=nil;
		LootCount_Chat("DropCount: Data-sharing OFF");
		return;
	end

	if (msg=="debug") then
		if (LootCount_DropCount_Debug) then
			LootCount_DropCount_Debug=nil;
			LootCount_Chat("DropCount: Debug: OFF");
		else
			LootCount_DropCount_Debug=true;
			LootCount_Chat("DropCount: Debug: ON");
		end
		return;
	end

	LootCount_Chat(LOOTCOUNT_DROPCOUNT_VERSIONTEXT);
	LootCount_DropCount_CleanDB();
	LootCount_Chat("DC: -> "..SLASH_DROPCOUNT2.." guild -> Share data with guild");
	LootCount_Chat("DC: -> "..SLASH_DROPCOUNT2.." raid -> Share data with party/raid");
	LootCount_Chat("DC: -> "..SLASH_DROPCOUNT2.." noshare -> Do not share data");
	if (LootCount_DropCount_DB.GUILD) then LootCount_Chat("DC: Currently sharing data with guild");
	elseif (LootCount_DropCount_DB.RAID) then LootCount_Chat("DC: Currently sharing data with party/raid");
	else LootCount_Chat("DC: Currently not sharing data"); end
end


-- An event has been received
function LootCount_DropCount_OnEvent(event)
	if (event=="COMBAT_LOG_EVENT_UNFILTERED") then
		if (arg2=="PARTY_KILL" and (bit.band(arg3,COMBATLOG_OBJECT_TYPE_PET) or bit.band(arg3,COMBATLOG_OBJECT_TYPE_PLAYER))) then
			if (GetNumPartyMembers()<1) then
				LootCount_DropCount_AddKill(true,arg6,arg7,LootCount_DropCount_Character.Skinning);
			end
			LootCountAPI.Force(LOOTCOUNT_DROPCOUNT);
		end
		return;
	end

	if (event=="PLAYER_FOCUS_CHANGED" or event=="PLAYER_TARGET_CHANGED") then
		local targettype=LootCount_DropCount_GetTargetType();
		if (not targettype) then
			LootCount_DropCount_Target=nil;
			LootCount_DropCount_SkinTarget=nil;
			LootCount_DropCount_SkinTargetKill=nil;
			return;
		end
		LootCount_DropCount_Target=nil;
		LootCount_DropCount_LastLootID=nil;
		if (not UnitIsDead(targettype)) then return; end
		if (UnitIsFriend("player",targettype)) then return; end
		LootCount_DropCount_GUID=UnitGUID(targettype);
		LootCount_DropCount_SkinTarget=UnitName(targettype);				-- Get current valid target
		LootCount_DropCount_SkinTargetKill=LootCount_DropCount_SkinTarget;	-- Set unit for skinning-drop
		if (UnitIsTapped(targettype)) and (not UnitIsTappedByPlayer(targettype)) then return; end	-- Not my kill (in case of skinning)
		LootCount_DropCount_Target=LootCount_DropCount_SkinTarget;			-- Save name of dead targetted/focused enemy
		LootCount_DropCount_LastLootID=time();
		return;
	end

	if (event=="LOOT_OPENED") then
		LootCount_DropCount_LootList=nil; LootCount_DropCount_LootList={};
		local slots=GetNumLootItems();
		if (slots<1) then return; end
		local skinning=true;
		for i=1,slots do
			LootCount_DropCount_LootList[i]={};
			_,_,LootCount_DropCount_LootList[i].Count=GetLootSlotInfo(i);	-- Returns icon path, item name, and item quantity for the item in the given loot window slot
			LootCount_DropCount_LootList[i].Item=LootCount_GetID(GetLootSlotLink(i));	-- Returns an itemLink for the item in the given loot window slot
			if (skinning and not LootCount_DropCount_LootIsSkinning(LootCount_DropCount_LootList[i].Item)) then skinning=nil; end
		end
		local mTable=LootCount_DropCount_LootedMobs;
		if (skinning) then mTable=LootCount_DropCount_SkinnedMobs; end	-- Not kills, but skinnings
		if (mTable[LootCount_DropCount_GUID]) then return; end			-- Loot already done for this one
		if (not skinning and LootCount_DropCount_Target) then			-- If not my kill
			LootCount_DropCount_AddKill(true,LootCount_DropCount_GUID,LootCount_DropCount_Target);
		end

		local now=time();
		-- Save loot
		mTable[LootCount_DropCount_GUID]=now;							-- Set it
		for i=1,slots do
			if (LootCount_DropCount_LootList[i].Count>0) then			-- Not money
				if (LootCount_DropCount_Target) then
					LootCount_DropCount_AddLocalLoot(LootCount_DropCount_GUID,LootCount_DropCount_Target,LootCount_DropCount_LootList[i].Item,LootCount_DropCount_LootList[i].Count);
				elseif (LootCount_DropCount_SkinTarget) then
					LootCount_DropCount_AddLocalLoot(LootCount_DropCount_GUID,LootCount_DropCount_SkinTarget,LootCount_DropCount_LootList[i].Item,LootCount_DropCount_LootList[i].Count);
				end
			end
		end
		-- Remove old mobs
		for guid,when in pairs(mTable) do
			if (now-when>LCDC_LOOTEDAGE) then mTable[guid]=nil; end
		end
		return;
	end

	if (event=="CHAT_MSG_ADDON") then
		if (arg1~=LCDC_PREFIX) then return; end
		if (LootCount_DropCount_DB.RAID) then
			if (arg3~="RAID" and arg3~="PARTY") then return; end
		elseif (LootCount_DropCount_DB.GUILD) then
			if (arg3~="GUILD") then return; end
		else return; end

		if (arg4==UnitName("player")) then return; end
		LootCount_DropCount_ParseMessage(arg2,arg4);
	end

	if (event=="VARIABLES_LOADED") then
		if (LootCount_DropCount_Character.ShowZoneMobs==nil) then LootCount_DropCount_Character.ShowZoneMobs=false; end
		if (LootCount_DropCount_Character.ShowZone==nil) then LootCount_DropCount_Character.ShowZone=true; end
		if (not LootCount_DropCount_DB.CHANNELSET) then
			LootCount_DropCount_DB.CHANNELSET=true;
		end
	end
end


function LootCount_DropCount_TransmitData(GUID,mob,item,count)
	local channel;
	if (LootCount_DropCount_DB.GUILD) then channel="GUILD";
	elseif (LootCount_DropCount_DB.RAID) then channel="RAID";
	else return; end

	local text;
	if (item and count) then
		text=LCDC_MOBLOOT..LCDC_SEPARATOR..GUID..LCDC_SEPARATOR..mob..LCDC_SEPARATOR..item..LCDC_SEPARATOR..count;
	else
		text=LCDC_MOBKILL..LCDC_SEPARATOR..GUID..LCDC_SEPARATOR..mob;
	end

	SendAddonMessage(LCDC_PREFIX,text,channel);
end


-- LCDC_MOBKILL - GUID - MOBNAME - ZONE
-- LCDC_MOBLOOT - GUID - MOBNAME - ITEM - COUNT
function LootCount_DropCount_ParseMessage(text,sender)
	local header,guid,mob,item,count=strsplit(LCDC_SEPARATOR,text);
	if (header==LCDC_MOBKILL) then
		LootCount_DropCount_AddKill(nil,guid,mob,nil,nil,true,item);
		if (LootCount_DropCount_Debug) then
			LootCount_Chat("Guildie kill: \'"..mob.."\'",0,1,0);
		end
	elseif (header==LCDC_MOBLOOT) then
		if (not count or not item) then return; end
		if (LootCount_DropCount_LootIsSkinning(item)) then
			if (not LootCount_DropCount_AlreadyReceivedSkin(guid)) then 
				LootCount_DropCount_SkinTargetKill=mob;
			end
		end
		LootCount_DropCount_AddLoot(guid,mob,item,count,true);
		local itemname=GetItemInfo(item);
		if (not itemname) then _,itemname=LootCount_GetID(item); end
		if (LootCount_DropCount_Debug) then
			LootCount_Chat("Guildie drop: \'"..mob.."\' -> \'"..itemname.."\'x"..count,0,1,0);
		end
	end
end


function LootCount_DropCount_AlreadyReceivedSkin(GUID)
	if (LootCount_DropCount_ReceivedSkins.A~=GUID and
		LootCount_DropCount_ReceivedSkins.B~=GUID and
		LootCount_DropCount_ReceivedSkins.C~=GUID) then
		LootCount_DropCount_ReceivedSkins.C=LootCount_DropCount_ReceivedSkins.B;
		LootCount_DropCount_ReceivedSkins.B=LootCount_DropCount_ReceivedSkins.A;
		LootCount_DropCount_ReceivedSkins.A=GUID;
		return nil;
	end
	return true;
end


function LootCount_DropCount_GetZone()
	local here=GetRealZoneText();			-- Set zone for last kill
	if (not here or here=="") then return nil; end
	local ss=GetSubZoneText();										-- Set subzone/area for last kill
	if (ss=="") then ss=nil; end
	if (ss) then here=here.." - "..ss; end
	return here;
end


function LootCount_DropCount_AddKill(oma,GUID,mob,reservedvariable,noadd,notransmit,otherzone)
	-- Check if already counted
	local i=LootCount_DropCount_QueueSize;
	while (i>0) do
		if (LootCount_DropCount_TimedQueue[i].GUID==GUID) then return; end
		i=i-1;
	end

	local now=time();
	if (not LootCount_DropCount_DB.Count[mob]) then LootCount_DropCount_DB.Count[mob]={ Kill=0 }; end
	if (not otherzone) then otherzone=LootCount_DropCount_GetZone(); end
	LootCount_DropCount_DB.Count[mob].Zone=otherzone;		-- Set zone for last kill
	if (not noadd) then
		LootCount_DropCount_DB.Count[mob].Kill=LootCount_DropCount_DB.Count[mob].Kill+1;
		if (not notransmit) then LootCount_DropCount_TransmitData(GUID,mob); end
	else
		return;
	end

	LootCount_DropCount_QueueSize=LootCount_DropCount_QueueSize+1;
	LootCount_DropCount_TimedQueue[LootCount_DropCount_QueueSize]={};
	i=LootCount_DropCount_QueueSize;
	while (i>1) do
		LootCount_DropCount_TimedQueue[i].Mob=LootCount_DropCount_TimedQueue[i-1].Mob;
		LootCount_DropCount_TimedQueue[i].GUID=LootCount_DropCount_TimedQueue[i-1].GUID;
		LootCount_DropCount_TimedQueue[i].Oma=LootCount_DropCount_TimedQueue[i-1].Oma;
		LootCount_DropCount_TimedQueue[i].Time=LootCount_DropCount_TimedQueue[i-1].Time;
		i=i-1;
	end
	LootCount_DropCount_TimedQueue[1].Mob=mob;
	LootCount_DropCount_TimedQueue[1].GUID=GUID;
	LootCount_DropCount_TimedQueue[1].Oma=oma;
	LootCount_DropCount_TimedQueue[1].Time=now;
end


function LootCount_DropCount_LootIsSkinning(item)
	if (type(item)=="string") then _,item=LootCount_GetID(item); end
	return LootCount_DropCount_SkinningList[item];
end


function LootCount_DropCount_AddLocalLoot(GUID,mob,item,count,notransmit)
	if (not LootCount_DropCount_Character.Skinning) then
		LootCount_DropCount_Character.Skinning=LootCount_DropCount_LootIsSkinning(item);			-- Update the skinning-flag
	end
	LootCount_DropCount_AddLoot(GUID,mob,item,count,notransmit,LootCount_DropCount_Character.Skinning);
end


function LootCount_DropCount_AddLootMob(GUID,mob,item)
	if (not LootCount_DropCount_DB.Item[item].Name) then LootCount_DropCount_DB.Item[item].Name={}; end
	-- New stuff, so make database ready
	if (not LootCount_DropCount_DB.Item[item].Name[mob]) then		-- Mob not in drop-list
		LootCount_DropCount_DB.Item[item].Name[mob]=0;				-- Not looted, but make entry (will be added later)
	end
	if (not LootCount_DropCount_DB.Count[mob]) then					-- Mob not in kill-list
		LootCount_DropCount_AddKill(nil,GUID,mob,nil,true,true,"");	-- Add it with zero kills
	end
end


-- "isskinner" is 'true' if the loot originator is a skinner, and 'nil' if not.
-- The origin of the data can be both communication and own loot.
--
-- As of now, this function does not use this information
function LootCount_DropCount_AddLoot(GUID,mob,item,count,notransmit,isskinner)
	if (not notransmit) then LootCount_DropCount_TransmitData(GUID,mob,item,count); end
	local now=time();
	local skinning=LootCount_DropCount_LootIsSkinning(item);
	if (not LootCount_DropCount_DB.Item[item]) then LootCount_DropCount_DB.Item[item]={ Name={}, Skinning=skinning; }; end
	LootCount_DropCount_DB.Item[item].Time=now;						-- Last point in time for loot of this item
	LootCount_DropCount_AddLootMob(GUID,mob,item,skinning);				-- Make register

	LootCount_DropCount_DB.Item[item].Name[mob]=LootCount_DropCount_DB.Item[item].Name[mob]+count;

	if (skinning) then		-- Skinner-loot, so add it as a skinning-kill
		if (LootCount_DropCount_SkinTargetKill and LootCount_DropCount_SkinTargetKill==mob) then
			if (not LootCount_DropCount_DB.Count[mob].Skinning) then LootCount_DropCount_DB.Count[mob].Skinning=0; end
			LootCount_DropCount_DB.Count[mob].Skinning=LootCount_DropCount_DB.Count[mob].Skinning+1;
			LootCount_DropCount_SkinTargetKill=nil;					-- Added, so next loot on this target is more than one items from same skinning
		end
	end
	LootCountAPI.Force(LOOTCOUNT_DROPCOUNT);
end


function LootCount_DropCount_GetRatio(item,mob)
	if (LCT_QUEST) then
		local _,_,_,_,_,itemtype=GetItemInfo(item);
		local _,itemID=LootCount_GetID(item);
		if (itemtype and itemtype==LCT_QUEST and not LootCount_DropCount_NoQuest[itemID]) then return LCR_QUEST; end
	end
	local kills=0;
	if (not LootCount_DropCount_DB.Item[item]) then return 0; end
	if (not LootCount_DropCount_DB.Item[item].Name) then return 0; end
	if (not LootCount_DropCount_DB.Item[item].Name[mob]) then return 0; end
	if (not LootCount_DropCount_DB.Count[mob]) then return 0; end
	if (not LootCount_DropCount_DB.Item[item].Skinning) then
		kills=LootCount_DropCount_DB.Count[mob].Kill;
	else
		kills=LootCount_DropCount_DB.Count[mob].Skinning;
	end
	if (not kills or kills<1) then return 0; end
	local drops=LootCount_DropCount_DB.Item[item].Name[mob];
	local nosafe; if (drops<2) then nosafe=true; end

	return drops/kills,nosafe;
end


function LootCount_DropCount_UpdateButton(button)
	if (not button) then return; end			-- End of iteration

	if (not button.User or not button.User.Texture) then return; end
	local texture=getglobal(button:GetName().."IconTexture");
	texture:SetTexture(button.User.Texture);			-- Set texture from item
	if (not LootCount_DropCount_DB.Item[button.User.itemID]) then return; end			-- Nothing assigned yet
	if (not LootCount_DropCount_DB.Item[button.User.itemID].Name) then return; end		-- No known droppers

	local inqueue={};
	local i=1;
	while (i<=LootCount_DropCount_QueueSize and LootCount_DropCount_TimedQueue[i]) do
		if (LootCount_DropCount_TimedQueue[i].Oma) then
			if (not inqueue[LootCount_DropCount_TimedQueue[i].Mob]) then
				inqueue[LootCount_DropCount_TimedQueue[i].Mob]={ Count=1, Ratio=LootCount_DropCount_GetRatio(button.User.itemID,LootCount_DropCount_TimedQueue[i].Mob) };
			else
				inqueue[LootCount_DropCount_TimedQueue[i].Mob].Count=inqueue[LootCount_DropCount_TimedQueue[i].Mob].Count+1;
			end
		end
		i=i+1;
	end

	local ratio=0;
	local count=0;
	for mob,mTable in pairs(inqueue) do
		count=count+mTable.Count;
		ratio=ratio+(mTable.Count*mTable.Ratio);
	end
	if (count<1) then return; end
	ratio=ratio/count;

	local goalvalue=nil;
	if (button.goal and button.goal>0) then
		local amount=LootCount_GetItemCount(button.User.itemID);
		if (amount>=button.goal) then goalvalue="OK";
		else
			goalvalue=button.goal-amount;
			if (ratio>0) then goalvalue=math.ceil(goalvalue/ratio);
			else goalvalue=""; end
		end
	end

	LootCountAPI.SetData(LOOTCOUNT_DROPCOUNT,button,LootCount_DropCount_FormatPC(ratio),goalvalue);
end


function LootCount_DropCount_FormatPC(ratio,addition)
	if (ratio<0) then return "Quest"; end
	if (not addition) then addition=""; end
	local text;
	local pc=ratio*100;
	if (pc>=100) then text=string.format("%.0f",pc);
	elseif (pc>=10) then text=string.format("%.0f",pc);
	elseif (pc>=1) then text=string.format("%.1f",pc);
	elseif (pc==0) then return "0"..addition;
	else text=string.format("%.02f",pc);
	end
	return text..addition;
end


function LootCount_DropCount_MakeRegister(button,itemID,clearit)
	if (not button.User) then button.User = { };
	elseif (clearit) then LootCount_ClearTable(button.User);
	end
	if (not button.User.itemID or itemID) then
		if (not itemID) then return nil; end
		button.User.itemID=itemID;
	end
	return true;
end


function LootCount_DropCount_DropItem(button,itemID)
	LootCount_DropCount_MakeRegister(button,itemID,true);						-- Start counting from now
	_,_,_,_,_,_,_,_,_,button.User.Texture=GetItemInfo(itemID);					-- Set custom texture
end


function LootCount_DropCount_GetTargetType()
	local targettype="playertarget";
	local mobname=UnitName(targettype);
	if (not mobname) then
		mobname=UnitName("focus"); if (not mobname) then return nil; end
		targettype="focus";
	end
	return targettype;
end


function LootCount_DropCount_IconClicked(button,LR,count)
	if (count~=1) then return; end
	if (LR=="RightButton") then
		LootCount_DropCount_ToggleMenu(button);
	end
end


function LootCount_DropCount_GetColour(ratio)
	if (ratio>1) then ratio=1; elseif (ratio<0) then ratio=0; end
	ratio=string.format("|cFF%02X%02X%02X",128+(ratio*127),128+(ratio*127),128+(ratio*127));		-- AARRGGBB
	return ratio;
end


function LootCount_DropCount_Tooltip(button)
	if (type(button)=="string") then
		button={
			FreeFloat=true;
			User={
				itemID=button;
			};
		};
	elseif (not button.User or not button.User.itemID) then
		GameTooltip:SetOwner(button,"ANCHOR_RIGHT");
		GameTooltip:SetText("Drop an item here");
		GameTooltip:Show();
		return;
	end
	if (not LootCount_DropCount_DB.Item[button.User.itemID]) then return; end
	local itemname,_,rarity=GetItemInfo(button.User.itemID);
	if (not itemname or not rarity) then LootCount_DropCount_StartCache(button.User.itemID) return; end
	local _,_,_,colour=GetItemQualityColor(rarity);
	if (button.FreeFloat) then GameTooltip:SetOwner(UIParent,"ANCHOR_CURSOR");
	else GameTooltip:SetOwner(button,"ANCHOR_RIGHT"); end
	GameTooltip:SetText(colour.."["..itemname.."]|r:");
	local currentzone=GetRealZoneText();
	if (not currentzone) then
		currentzone="";
	elseif (LootCount_DropCount_Character.ShowZoneMobs) then
		GameTooltip:AddDoubleLine("Showing mobs from "..currentzone.." only","",0,1,1,0,1,1);
	end
	local skinningdrop=LootCount_DropCount_DB.Item[button.User.itemID].Skinning;
	if (skinningdrop) then GameTooltip:AddDoubleLine("Skinning-drop","",1,0,1,1,0,1); end

	local list={};
	local line=1;
	local droplist=0;
	for mob,drops in pairs(LootCount_DropCount_DB.Item[button.User.itemID].Name) do
		local pretext="";
		local i=1;
		local Show=nil;
		while(LootCount_DropCount_TimedQueue[i]) do
			if (LootCount_DropCount_TimedQueue[i].Mob==mob and LootCount_DropCount_TimedQueue[i].Oma) then
				pretext="-> ";
				Show=true;						-- Show this one no matter what
				droplist=droplist+1;			-- Count a hi-pri entry
				i=LootCount_DropCount_QueueSize;
			end
			i=i+1;
		end

		local low=64;
		local high=255;
		list[line]={};

		if ((not skinningdrop) and LootCount_DropCount_DB.Count[mob].Kill) then
			list[line].Count=LootCount_DropCount_DB.Count[mob].Kill;
		elseif (skinningdrop and LootCount_DropCount_DB.Count[mob].Skinning) then
			list[line].Count=LootCount_DropCount_DB.Count[mob].Skinning;
		else
			list[line].Count=0;
		end
		local saturation=((high-low)/(LCDC_KILLS_SAFE-LCDC_KILLS_HIDE))*list[line].Count;
		if (saturation<0) then saturation=0;
		elseif (saturation>(high-low)) then saturation=(high-low); end
		local colour=string.format("|cFF%02X%02X%02X",high-saturation,low+saturation,0);			-- AARRGGBB
		list[line].ratio=LootCount_DropCount_GetRatio(button.User.itemID,mob);

		local zone="";
		if (LootCount_DropCount_Character.ShowZone and LootCount_DropCount_DB.Count[mob].Zone) then
			zone=" |cFF0060FF("..LootCount_DropCount_DB.Count[mob].Zone..")";
		end

		list[line].Ltext=colour..pretext..mob..zone.."|r: ";
		list[line].Show=Show;
		if (LootCount_DropCount_Character.ShowZoneMobs and LootCount_DropCount_DB.Count[mob].Zone and currentzone and string.find(LootCount_DropCount_DB.Count[mob].Zone,currentzone)~=1) then
			list[line]=nil;
			line=line-1;
		end
		line=line+1;
	end
	list=LootCount_DropCount_SortRatio(list);

	-- Type list
	local count=0;
	local supressed=0;
	local goal=LCDC_LISTLENGTH-droplist;						-- Subtract hi-pri entry
	line=1;
	while(list[line]) do
		if ((count<goal and list[line].Count>=LCDC_KILLS_HIDE) or list[line].Show) then
			GameTooltip:AddDoubleLine(list[line].Ltext,list[line].Rtext,1,1,1,1,1,1);
			if (not list[line].Show) then count=count+1; end	-- Count all normal entries
		else
			supressed=supressed+1;
		end
		line=line+1;
	end
	if (supressed>0) then
		GameTooltip:AddDoubleLine(supressed.." more entries","",1,.5,1,1,1,1);
	end

	GameTooltip:Show();
end


function LootCount_DropCount_SortRatio(list)
	local line=1;
	while(list[line+1]) do
		if (list[line].ratio<list[line+1].ratio) then
			list[line].ratio,list[line+1].ratio=list[line+1].ratio,list[line].ratio;
			list[line].Ltext,list[line+1].Ltext=list[line+1].Ltext,list[line].Ltext;
			list[line].Show,list[line+1].Show=list[line+1].Show,list[line].Show;
			list[line].Count,list[line+1].Count=list[line+1].Count,list[line].Count;
			list[line].NoSafe,list[line+1].NoSafe=list[line+1].NoSafe,list[line].NoSafe;
			if (line>1) then line=line-2; end
		end
		line=line+1;
	end
	local line=1;
	while(list[line]) do
		local addin=LootCount_DropCount_FormatPC(list[line].ratio,"%|r");
		if (list[line].NoSafe) then addin="<"..addin; end
		list[line].Rtext=LootCount_DropCount_GetColour(list[line].ratio)..addin;
		line=line+1;
	end
	return list;
end


function LootCount_DropCount_SetLootlist(unit)
	if (not LootCount_DropCount_DB.Count[unit]) then return; end

	LootCount_DropCount_TT:ClearLines();
	LootCount_DropCount_TT:SetOwner(UIParent,"ANCHOR_CURSOR");
	LootCount_DropCount_TT:SetText(unit);
	local text="";
	if (LootCount_DropCount_DB.Count[unit].Skinning) then text="(Skinned "..LootCount_DropCount_DB.Count[unit].Skinning.." times)"; end
	LootCount_DropCount_TT:AddDoubleLine(LootCount_DropCount_DB.Count[unit].Kill.." kills",text,.4,.4,1,1,0,1);

	local list={};
	local line=1;
	local missingitems=nil;
	local itemsinlist=nil;
	for item,iTable in pairs(LootCount_DropCount_DB.Item) do
		for mob,mTable in pairs(iTable.Name) do
			if (mob==unit) then
				local itemname,_,rarity=GetItemInfo(item);
				if (not itemname or not rarity) then
					LootCount_DropCount_StartCache(item);
					missingitems=true;
				else
					local _,_,_,colour=GetItemQualityColor(rarity);
					local thisratio,thissafe=LootCount_DropCount_GetRatio(item,mob);
					list[line]={ Ltext=colour.."["..itemname.."]|r: ", ratio=thisratio, NoSafe=thissafe };
					if (LootCount_DropCount_DB.Item[item].Skinning) then
						list[line].Ltext=list[line].Ltext.."|cFFFF00FF*|r";							-- AARRGGBB
					end
					line=line+1;
					itemsinlist=true;
				end
			end
		end
	end
	if (missingitems) then
		LootCount_DropCount_TT:AddDoubleLine("Missing some items.","",1,0,0,0,0,0);
		LootCount_DropCount_TT:AddDoubleLine("Loading...","",1,0,0,0,0,0);
		LootCount_DropCount_TT:Show();
		LootCount_DropCount_TT.Loading=true;
		return;
	end
	LootCount_DropCount_TT.Loading=nil;
	if (not itemsinlist) then LootCount_DropCount_TT:Hide(); return; end

	list=LootCount_DropCount_SortRatio(list);
	line=1;
	while(list[line]) do
		LootCount_DropCount_TT:AddDoubleLine(list[line].Ltext,list[line].Rtext,1,1,1,1,1,1);
		line=line+1;
	end
	LootCount_DropCount_TT:Show();
end


function LootCount_DropCount_StartCache(item)
	LootCount_DropCount_CacheRetries=0;
	if (not LootCount_DropCount_UnknownItems) then LootCount_DropCount_UnknownItems={}; LootCount_DropCount_EnumerateTimer=.5; end
	LootCount_DropCount_UnknownItems[item]=true;
end


function LootCount_DropCount_CacheItem(item)
	if (type(item)=="number") then item="item:"..item;
	elseif (type(item)~="string") then return true; end
	local name=GetItemInfo(item);
	if (not LootCount_DropCount_CF:IsVisible()) then
		if (not name) then
			LootCount_DropCount_CF:SetOwner(UIParent); LootCount_DropCount_CF:SetHyperlink(item); LootCount_DropCount_CF:Hide();
			LootCount_DropCount_CacheRetries=LootCount_DropCount_CacheRetries+1;
			return false;
		else
			LootCount_DropCount_CacheRetries=0;
		end
	end
	return true;
end


function LootCount_DropCount_ToggleMenu(button)
	LootCount_DropCount_This=button;
	ToggleDropDownMenu(nil,nil,LootCount_DropCount_MenuOptions,button,0,0);
end


function LootCount_DropCount_BuildMenu(button)
	LootCount_DropCount_This=button;
	UIDropDownMenu_Initialize(LootCount_DropCount_MenuOptions,LootCount_DropCount_MenuLoad,"MENU");
end


function LootCount_DropCount_MenuLoad()
	local info={};
	info.textR=1; info.textG=1; info.textB=1;
	if (LCT_QUEST and LootCount_DropCount_This.User and LootCount_DropCount_This.User.itemID) then
		local _,_,_,_,_,itemtype=GetItemInfo(LootCount_DropCount_This.User.itemID);
		if (itemtype and itemtype==LCT_QUEST) then
			info.isTitle=1; info.checked=nil; info.func=nil; info.text="Setting for this item:"; UIDropDownMenu_AddButton(info,1); info.isTitle=nil; info.disabled=nil;
			local _,thisid=LootCount_GetID(LootCount_DropCount_This.User.itemID);
			if (not LootCount_DropCount_NoQuest[thisid]) then info.checked=true; end
			info.text="Only drops when on a quest"; info.value=thisid; info.func=LootCount_DropCount_Menu_ToggleQuestItem; UIDropDownMenu_AddButton(info,1);

			info.isTitle=1; info.checked=nil; info.func=nil; info.text=" "; UIDropDownMenu_AddButton(info,1);
		end
	end

	info.isTitle=1; info.checked=nil; info.func=nil; info.text="Global settings for "..LOOTCOUNT_DROPCOUNT_VERSIONTEXT..":"; UIDropDownMenu_AddButton(info,1); info.isTitle=nil; info.disabled=nil;

	info.func=LootCount_DropCount_Menu_ToggleZone;
	info.text="Show zone"; info.checked=LootCount_DropCount_Character.ShowZone; UIDropDownMenu_AddButton(info,1);
	info.func=LootCount_DropCount_Menu_ToggleZoneMobs;
	info.text="Only list mobs from current zone"; info.checked=LootCount_DropCount_Character.ShowZoneMobs; UIDropDownMenu_AddButton(info,1);
	info.func=LootCount_DropCount_Menu_SetGoal;
	info.text="Set goal"; info.checked=nil; UIDropDownMenu_AddButton(info,1);

	info.func=LootCount_DropCount_Menu_ToggleChannel;
	info.text="Guild data-sharing"; info.value="GUILD"; info.checked=LootCount_DropCount_DB[info.value]; UIDropDownMenu_AddButton(info,1);
	info.text="Group/raid data-sharing"; info.value="RAID"; info.checked=LootCount_DropCount_DB[info.value]; UIDropDownMenu_AddButton(info,1);
	info.checked=nil;
end


function LootCount_DropCount_Menu_SetGoal()
	LootCount_SetGoalPopup(LootCount_DropCount_This);
end


function LootCount_DropCount_Menu_ToggleZoneMobs()
	if (LootCount_DropCount_Character.ShowZoneMobs) then LootCount_DropCount_Character.ShowZoneMobs=false;
	else LootCount_DropCount_Character.ShowZoneMobs=true; end
end


function LootCount_DropCount_Menu_ToggleZone()
	if (LootCount_DropCount_Character.ShowZone) then LootCount_DropCount_Character.ShowZone=false;
	else LootCount_DropCount_Character.ShowZone=true; end
end


function LootCount_DropCount_Menu_ToggleQuestItem()
	if (LootCount_DropCount_NoQuest[this.value]) then LootCount_DropCount_NoQuest[this.value]=nil; return; end
	LootCount_DropCount_NoQuest[this.value]=true;
end


function LootCount_DropCount_Menu_ToggleChannel()
	if (LootCount_DropCount_DB[this.value]) then LootCount_DropCount_DB[this.value]=nil; else LootCount_DropCount_DB[this.value]=true; end
	if (this.value=="GUILD") then LootCount_DropCount_DB.RAID=nil; end
	if (this.value=="RAID") then LootCount_DropCount_DB.GUILD=nil; end
end


function LootCount_DropCount_Menu_SetGoal()
	local user=LootCountAPI.User(LootCount_DropCount_This);
	LootCount_SetGoalPopup(LootCount_DropCount_This);
end


function LootCount_DropCount_CleanDB()
	local text="DropCount: ";

	if (LootCount_DropCount_Debug) then
		collectgarbage("collect");
		UpdateAddOnMemoryUsage();
		local usage=GetAddOnMemoryUsage("LootCount_DropCount");
		usage=string.format("%.0fKB",usage);
		text="DropCount memory usage: "..usage.." -> ";
	end

	local nowitems,nowmobs;
	nowitems=0;
	if (LootCount_DropCount_DB.Item) then
		for item,iTable in pairs(LootCount_DropCount_DB.Item) do nowitems=nowitems+1; end
		text=text..nowitems.." items";
	end
	if (not LootCount_DropCount_StartItems) then LootCount_DropCount_StartItems=nowitems; end
	nowmobs=0;
	if (LootCount_DropCount_DB.Count) then
		for creature,cTable in pairs(LootCount_DropCount_DB.Count) do nowmobs=nowmobs+1; end
		text=text.." -> "..nowmobs.." creatures";
	end
	if (not LootCount_DropCount_StartMobs) then LootCount_DropCount_StartMobs=nowmobs; end

	LootCount_Chat(text);
	LootCount_Chat("DropCount: New this session: "..nowitems-LootCount_DropCount_StartItems.." items, "..nowmobs-LootCount_DropCount_StartMobs.." mobs");
end


function LootCount_DropCount_OnUpdate(elapsed)
	if (not LCT_QUEST and (not LootCount_DropCount_UnknownItems or not LootCount_DropCount_UnknownItems["item:31812"])) then
		_,_,_,_,_,LCT_QUEST=GetItemInfo("item:31812");
		if (not LCT_QUEST) then
			LootCount_DropCount_StartCache("item:31812")
		end
	end

	if (LootCount_DropCount_UnknownItems) then
		LootCount_DropCount_EnumerateTimer=LootCount_DropCount_EnumerateTimer-elapsed;
		if (LootCount_DropCount_EnumerateTimer<=0) then
			if (not LootCount_DropCount_RunningItem) then
				for unknown,value in pairs(LootCount_DropCount_UnknownItems) do LootCount_DropCount_RunningItem=unknown; end
				if (not LootCount_DropCount_RunningItem) then LootCount_DropCount_UnknownItems=nil; return; end
			end
			local itemname=GetItemInfo(LootCount_DropCount_RunningItem);
			if (not itemname) then
				LootCount_DropCount_CacheItem(LootCount_DropCount_RunningItem);
				LootCount_DropCount_EnumerateTimer=2.5;
			else
				LootCount_DropCount_UnknownItems[LootCount_DropCount_RunningItem]=nil;
				LootCount_DropCount_RunningItem=nil;
				LootCount_DropCount_CacheRetries=0;
			end
		end
		if (LootCount_DropCount_CacheRetries>=LCSV_MAXRETRIES) then
			LootCount_DropCount_UnknownItems=nil;			-- Too many tries, so abort
			LootCount_DropCount_RunningItem=nil;
		end
	end

	if (LootCount_DropCount_Registered) then
		if (not LootCount_InCombat and LootCount_DropCount_Fighting) then
			LootCount_DropCount_Fighting=nil;
		elseif (LootCount_InCombat and not LootCount_DropCount_Fighting) then
			LootCount_DropCount_Fighting=true;
		end

		if (IsAltKeyDown() and UnitExists("mouseover")) then
			if (not LootCount_DropCount_TT:IsVisible()) then LootCount_DropCount_SetLootlist(UnitName("mouseover"));
			elseif (LootCount_DropCount_TT.Loading) then
				LootCount_DropCount_TT:Hide();
				LootCount_DropCount_SetLootlist(UnitName("mouseover"));
			end
		else
			if (LootCount_DropCount_TT:IsVisible()) then LootCount_DropCount_TT:Hide(); end
		end

		if (LootCount_DropCount_QueueSize>0) then
			local now=time();
			if (LootCount_DropCount_QueueSize>0 and now-LootCount_DropCount_TimedQueue[LootCount_DropCount_QueueSize].Time>LCDC_QUEUETIME) then
				LootCount_DropCount_TimedQueue[LootCount_DropCount_QueueSize]=nil;
				LootCount_DropCount_QueueSize=LootCount_DropCount_QueueSize-1;
				LootCountAPI.Force(LOOTCOUNT_DROPCOUNT);
			end
		end

		if (GameTooltip and IsAltKeyDown() and GameTooltip:IsVisible()) then
			local _,ThisItem=GameTooltip:GetItem();
			if (ThisItem) then
				ThisItem=LootCount_GetID(ThisItem);
				if (ThisItem) then
					LootCount_DropCount_Tooltip(ThisItem);
				end
			end
		end

		return;
	end

	if (not LootCountAPI or not LootCountAPI.Register or not LootCount_Loaded) then return; end
	
	LootCount_DropCount_CleanDB();
	local info = {	Name=LOOTCOUNT_DROPCOUNT,
					MenuText="LootCount Drop-rate",
					Update=LootCount_DropCount_UpdateButton,
					Drop=LootCount_DropCount_DropItem,
					Clicker=LootCount_DropCount_IconClicked,
					Texture="Interface\\Icons\\INV_Misc_QuestionMark",
					Tooltip=LootCount_DropCount_Tooltip,
					ColorBottom={ r=1, g=1, b=0 },
					ColorTop={ r=.7, g=.7, b=0 },
				};

	LootCount_DropCount_DB.RAID=nil;
	if (IsInGuild()) then LootCount_DropCount_DB.GUILD=true; else LootCount_DropCount_DB.GUILD=nil; end
	LootCountAPI.Register(info);
	LootCount_DropCount_Registered=true;
end
